pub use config::AdminModelConfig;
pub use dto::AdminApp;
pub use dto::AdminModel;
pub use repository::{
    AdminRepository, LookupKey, RepoInfo, RepositoryContext, RepositoryInfo, RepositoryItem,
    RepositoryList, Widget,
};

mod auth {

    struct AdminUser {}

    struct AdminRole {}

    struct AdminGroup {}

    struct AdminActionLog {}
}

mod config {
    // user uses this configuration object to register another model.
    pub struct AdminModelConfig {
        pub name: String,
        pub app_key: String,
    }
}

mod dto {
    use serde::{Deserialize, Serialize};

    #[derive(Deserialize, Serialize)]
    pub struct AdminModel {
        pub key: String,
        pub name: String,

        pub admin_url: String,
        pub view_only: bool,

        pub add_url: Option<String>,
    }

    #[derive(Deserialize, Serialize)]
    pub struct AdminApp {
        pub key: String,
        pub name: String,

        pub app_url: String,
        pub models: Vec<AdminModel>,
    }
}

pub mod repository {
    use super::dto::{AdminApp, AdminModel};
    use async_trait::async_trait;
    use derive_builder::Builder;
    use serde::{Serialize, Serializer};
    use serde_json::Value;
    use std::{collections::HashMap, vec::IntoIter};

    pub type RepositoryContext = AdminModel;

    impl RepositoryContext {
        pub fn build_item(&self, key: &str, fields: Value) -> RepositoryItem {
            RepositoryItem {
                detail_url: Some(format!("{}/detail/{}", self.admin_url, key)),
                fields: fields,
            }
        }
    }

    /// This is a static configuration object.
    /// It might be changed in the future to have a dynamic counterpart.
    ///
    /// ## Example:
    /// Creating a simple required and readonly text input field with a label:
    ///
    /// ```ignore
    /// let my_field = Field::widget("/admin/widgets/input_text.jinja")
    ///     .labelled("Username")
    ///     .required()
    ///     .readonly();
    /// ```
    #[derive(Debug, Serialize, Clone, Copy)]
    pub struct Widget {
        pub widget: &'static str,
        pub label: Option<&'static str>,
        #[serde(rename = "type")]
        pub field_type: &'static str,
        pub required: bool,
        pub readonly: bool,
        pub options: &'static [(&'static str, &'static str)],
    }

    impl Widget {
        pub fn widget(widget: &'static str) -> Self {
            Widget {
                widget: widget,
                label: None,
                field_type: "text",
                required: false,
                readonly: false,
                options: &[],
            }
        }

        pub fn default() -> Self {
            Self::widget("/admin/widgets/input_text.jinja")
        }

        pub fn textarea() -> Self {
            Self::widget("/admin/widgets/input_textarea.jinja")
        }

        pub fn required(mut self) -> Self {
            self.required = true;
            self
        }

        pub fn readonly(mut self) -> Self {
            self.readonly = true;
            self
        }

        pub fn labeled(mut self, label: &'static str) -> Self {
            self.label = Some(label);
            self
        }

        pub fn as_password(mut self) -> Self {
            self.field_type = "password";
            self
        }

        pub fn as_hidden(mut self) -> Self {
            self.field_type = "hidden";
            self
        }

        pub fn options(mut self, options: &'static [(&'static str, &'static str)]) -> Self {
            self.options = options;
            self
        }
    }

    #[derive(Debug, Serialize)]
    struct Field {
        widget: String,
        label: Option<String>,
        #[serde(rename = "type")]
        field_type: String,
        readonly: bool,
        required: bool,
        options: Value,
    }

    impl From<Widget> for Field {
        fn from(value: Widget) -> Self {
            Field {
                widget: value.widget.to_string(),
                label: value.label.map(|s| s.to_string()),
                field_type: value.field_type.to_string(),
                readonly: value.readonly,
                required: value.required,
                options: value
                    .options
                    .iter()
                    .map(|(key, val)| (key.to_string(), serde_json::json!(val)))
                    .collect::<serde_json::Map<String, Value>>()
                    .into(),
            }
        }
    }

    #[derive(PartialEq)]
    pub enum LookupKey {
        Integer(usize),
        String(String),
    }

    // Note that LookupKey auto converts to integer.
    impl From<String> for LookupKey {
        fn from(s: String) -> Self {
            if let Ok(int_key) = s.parse::<usize>() {
                LookupKey::Integer(int_key)
            } else {
                LookupKey::String(s)
            }
        }
    }

    impl From<&str> for LookupKey {
        fn from(s: &str) -> Self {
            if let Ok(int_key) = s.parse::<usize>() {
                LookupKey::Integer(int_key)
            } else {
                LookupKey::String(s.to_owned())
            }
        }
    }

    impl From<usize> for LookupKey {
        fn from(i: usize) -> Self {
            LookupKey::Integer(i)
        }
    }

    impl std::fmt::Display for LookupKey {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                LookupKey::Integer(i) => write!(f, "{}", i),
                LookupKey::String(s) => write!(f, "{}", s),
            }
        }
    }

    #[derive(Serialize)]
    pub struct RepositoryItem {
        pub fields: Value,
        pub detail_url: Option<String>,
    }

    pub enum RepositoryList {
        Empty,
        List {
            values: Vec<RepositoryItem>,
        },
        Page {
            values: Vec<RepositoryItem>,
            offset: usize,
            total: usize,
        },
        Stream {
            values: Vec<RepositoryItem>,
            next_index: Option<String>,
        },
    }

    impl IntoIterator for RepositoryList {
        type Item = RepositoryItem;
        type IntoIter = IntoIter<Self::Item>;

        fn into_iter(self) -> Self::IntoIter {
            match self {
                RepositoryList::Empty => vec![].into_iter(),
                RepositoryList::List { values } => values.into_iter(),
                RepositoryList::Page { values, .. } => values.into_iter(),
                RepositoryList::Stream { values, .. } => values.into_iter(),
            }
        }
    }

    impl Serialize for RepositoryList {
        fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
        where
            S: Serializer,
        {
            match self {
                RepositoryList::Empty => serializer.serialize_unit(),
                RepositoryList::List { values }
                | RepositoryList::Page { values, .. }
                | RepositoryList::Stream { values, .. } => values.serialize(serializer),
            }
        }
    }

    /// Static initializer for RepositoryInfo.
    pub struct RepoInfo {
        pub name: &'static str,
        pub lookup_key: &'static str,
        pub display_list: &'static [&'static str],
        pub fields: &'static [&'static str],
    }

    impl RepoInfo {
        pub fn build(self) -> RepositoryInfo {
            self.into()
        }
    }

    #[derive(Serialize)]
    pub struct RepositoryInfo {
        name: String,
        lookup_key: String,
        display_list: Vec<String>,
        fields: Vec<(String, Field)>,
    }

    impl RepositoryInfo {
        pub fn new(name: &str, lookup_key: &str) -> Self {
            RepositoryInfo {
                name: name.to_owned(),
                lookup_key: lookup_key.to_owned(),
                display_list: vec![],
                fields: vec![],
            }
        }

        // self mutating builder pattern
        pub fn display_list(mut self, display_list: &[&str]) -> Self {
            self.display_list = display_list.iter().map(|&e| e.to_string()).collect();
            self
        }

        // set field for a given key.
        pub fn set_widget(mut self, name: &str, widget: Widget) -> Self {
            // Find the index of the existing entry with the same name, if it exists
            let pos = self
                .fields
                .iter()
                .position(|(existing_name, _)| existing_name == name);

            match pos {
                Some(index) => {
                    // If found, replace the `Field` part of the tuple
                    self.fields[index].1 = Field::from(widget);
                }
                None => {
                    // If not found, add a new entry
                    self.fields.push((name.to_owned(), Field::from(widget)));
                }
            }
            self
        }
    }

    impl From<RepoInfo> for RepositoryInfo {
        fn from(repo_info: RepoInfo) -> Self {
            RepositoryInfo {
                name: repo_info.name.to_string(),
                lookup_key: repo_info.lookup_key.to_string(),
                display_list: repo_info
                    .display_list
                    .iter()
                    .map(|&s| s.to_string())
                    .collect(),
                fields: repo_info
                    .fields
                    .iter()
                    .map(|x| (x.to_string(), Field::from(Widget::default())))
                    .collect(),
            }
        }
    }

    #[async_trait]
    pub trait AdminRepository: Send + Sync {
        async fn info(&self, context: &RepositoryContext) -> RepositoryInfo;
        async fn list(&self, context: &RepositoryContext) -> RepositoryList;
        async fn get(&self, context: &RepositoryContext, id: LookupKey) -> Option<RepositoryItem>;
        async fn create(
            &mut self,
            context: &RepositoryContext,
            data: Value,
        ) -> Option<RepositoryItem>;
        async fn update(
            &mut self,
            context: &RepositoryContext,
            id: LookupKey,
            data: Value,
        ) -> Option<RepositoryItem>;
        async fn delete(&mut self, context: &RepositoryContext, id: LookupKey) -> Option<Value>;
    }
}
